from dessia_common.vectored_objects import Catalog, ParetoSettings
from dessia_common.core import DessiaFilter
import dessia_common.workflow as wf
import dessia_common.typings as dct

choice_args = ['MPG', 'Cylinders', 'Displacement', 'Horsepower',
               'Weight', 'Acceleration', 'Model']  # Ordered

minimized_attributes = {'MPG': False, 'Horsepower': True,
                        'Weight': True, 'Acceleration': False}

aimpoint = {'MPG': 4, 'Weight': 1000}

filters = [DessiaFilter(attribute='MPG',
                        operator='gte',
                        bound=10),
           DessiaFilter(attribute='MPG',
                        operator='lte',
                        bound=35)]

# Blocks
import_csv = wf.Import(type_='csv')
instantiate_pareto = wf.InstanciateModel(model_class=ParetoSettings,
                                         name='Pareto Settings')
instantiate_catalog = wf.InstanciateModel(model_class=Catalog,
                                          name='Cars Instantiation')
filter_method = wf.ModelMethod(dct.MethodType(Catalog, 'filter_'),
                               name='Filters')
filtered_catalog = wf.InstanciateModel(model_class=Catalog,
                                       name='Filtered Catalog')
display = wf.Display(order=0, name="Display")
# display = wf.MultiPlot(choice_args, 1, name='Display')
filtered = wf.Display(order=1, name="Filtered")
# filtered = wf.MultiPlot(choice_args, 1, name='Filtered')
# objectives_method = wf.ModelMethod(model_class=Catalog,
#                                    method_name='find_best_objective',
#                                    name="Find best objectives")
# objectives_attributes = wf.ModelAttribute(attribute_name='objectives',
#                                           name='Objectives')

blocks = [import_csv, instantiate_pareto, instantiate_catalog, filter_method,
          filtered_catalog, display, filtered]
# , objectives_method, objectives_attributes]

# Pipes
pipes = [
    wf.Pipe(import_csv.outputs[0], instantiate_catalog.inputs[0]),
    wf.Pipe(import_csv.outputs[1], instantiate_catalog.inputs[1]),
    wf.Pipe(instantiate_pareto.outputs[0], instantiate_catalog.inputs[2]),
    wf.Pipe(instantiate_catalog.outputs[0], display.inputs[0]),
    # wf.Pipe(instantiate_catalog.outputs[0], objectives_method.inputs[0]),
    wf.Pipe(instantiate_catalog.outputs[0], filter_method.inputs[0]),
    wf.Pipe(filter_method.outputs[0], filtered_catalog.inputs[0]),
    wf.Pipe(import_csv.outputs[1], filtered_catalog.inputs[1]),
    wf.Pipe(filtered_catalog.outputs[0], filtered.inputs[0]),
    wf.Pipe(instantiate_pareto.outputs[0], filtered_catalog.inputs[2])
    # wf.Pipe(objectives_method.outputs[1], objectives_attributes.inputs[0])
]

# Workflow
vectored_workflow = wf.Workflow(blocks=blocks, pipes=pipes,
                       output=filter_method.outputs[0],
                       name='Cars workflow')

# # Input values
input_values = {
    vectored_workflow.index(import_csv.inputs[0]): 'cars.csv',
    vectored_workflow.index(instantiate_pareto.inputs[0]): minimized_attributes,
    vectored_workflow.index(instantiate_pareto.inputs[1]): True,
    vectored_workflow.index(instantiate_catalog.inputs[4]): choice_args,
    vectored_workflow.index(instantiate_catalog.inputs[3]): [],
    vectored_workflow.index(instantiate_catalog.inputs[5]): 'Cars',
    vectored_workflow.index(filter_method.inputs[1]): filters,
    vectored_workflow.index(filtered_catalog.inputs[4]): choice_args,
    vectored_workflow.index(filtered_catalog.inputs[3]): [],
    vectored_workflow.index(filtered_catalog.inputs[5]): 'Filtered Cars',
    # workflow.index(objectives_method.inputs[1]): aimpoint,
    # workflow.index(objectives_method.inputs[2]): minimized_attributes
}
workflow_run = vectored_workflow.run(input_values=input_values)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 14 11:54:34 2019

@author: jezequel
"""
from powerpack import cells, power_profile, electrical
from powerpack.optimization import electrical as eo
from dessia_common import workflow as wf
from dessia_api_client import Client
import dessia_common as dc


# =============================================================================
# Vehicle specs
# =============================================================================
specs = {'SCx' : 1,
         'Crr' : 0.01,
         'mass' : 1200,
         'powertrain_efficiency' : 0.7,
         'charge_efficiency' : 0.5}

# =============================================================================
# Electrical Specs
# =============================================================================

limits_voltage_module = {'charge': {'minimum': 1, 'maximum': 100},
                         'discharge': {'minimum': 1, 'maximum': 120}}
limits_current_module = {'charge': {'minimum': 0, 'maximum': 100},
                         'discharge': {'minimum': -100, 'maximum': 0}}
limits_voltage_battery = {'charge': {'minimum': 0, 'maximum': 500},
                          'discharge': {'minimum': 0, 'maximum': 500}}
limits_current_battery = {'charge': {'minimum': 0, 'maximum': 1000},
                          'discharge': {'minimum': -1000, 'maximum': 0}}


power_profile_1, time_1 = power_profile.WltpProfile(specs, (1, ))
t_wltp1 = electrical.Evolution(list(time_1))
p_wltp1 = electrical.Evolution(list(power_profile_1))

power_profile_2, time_2 = power_profile.WltpProfile(specs, (2, ))
t_wltp2 = electrical.Evolution(list(time_2))
p_wltp2 = electrical.Evolution(list(power_profile_2))

power_profile_3, time_3 = power_profile.WltpProfile(specs, (3, ))
t_wltp3 = electrical.Evolution(list(time_3))
p_wltp3 = electrical.Evolution(list(power_profile_3))
t_load = electrical.Evolution(list(range(10)))
p_load = electrical.Evolution([1e5]*10)
t_end = electrical.Evolution(list(range(10)))
p_end = electrical.Evolution([-2e4]*10)

ce_end = electrical.CombinationEvolution(evolution1=[t_end],
                                   evolution2=[p_end],
                                   name='End Profile')
ce_wltp1 = electrical.CombinationEvolution(evolution1=[t_wltp1],
                                     evolution2=[p_wltp1],
                                     name='WLTP1 profile')
ce_wltp2 = electrical.CombinationEvolution(evolution1=[t_wltp2],
                                     evolution2=[p_wltp2],
                                     name='WLTP2 profile')
ce_wltp3 = electrical.CombinationEvolution(evolution1=[t_wltp3],
                                     evolution2=[p_wltp3],
                                     name='WLTP3 profile')
ce_load = electrical.CombinationEvolution(evolution1=[t_load],
                                    evolution2=[p_load],
                                    name='Load Profile')

load_bat = electrical.PowerProfile(soc_init=0.05*180000,
                             combination_evolutions=[ce_load],
                             loop=True,
                             soc_end=0.95*180000,
                             charger=True,
                             name='Load profile')
end_bat = electrical.PowerProfile(combination_evolutions=[ce_end],
                            loop=False,
                            power_accuracy=0.2,
                            soc_init=0.1*180000,
                            name='End profile')
wltp_bat = electrical.PowerProfile(combination_evolutions=[ce_wltp1, ce_wltp2, ce_wltp3],
                             loop=True,
                             power_accuracy=0.2,
                             soc_init=0.95*180000,
                             max_loop=1,
                             soc_end=0.1*180000,
                             use_selection=False,
                             name='WLTP profile')

comb_profile_wltp = electrical.CombinationPowerProfile([wltp_bat],
                                                 name='wltp_profil')
comb_profile_load = electrical.CombinationPowerProfile([load_bat],
                                                 name='load_profil')
comb_profile_end = electrical.CombinationPowerProfile([end_bat],
                                                name='end_soc_profil')

# =============================================================================
# Electrical Optimizer
# =============================================================================
input_values = {}
blocks = []
block_ebo = wf.InstanciateModel(eo.ElecBatteryOptimizer, name='EBO')
optimize_ebo = wf.ModelMethod(eo.ElecBatteryOptimizer, 'Optimize', 'Optimize EBO')
attribute_selection_ebo = wf.ModelAttribute('powerpack_electric_simulators', 'Attribute Selection EBO')

filters = [{'attribute' : 'bms.battery.number_module_parallel', 'operator' : 'gt', 'bound' : 2},
           {'attribute' : 'bms.battery.number_module_serie', 'operator' : 'lte', 'bound' : 10},
           {'attribute' : 'bms.number_cells', 'operator' : 'gte', 'bound' : 750},
           {'attribute' : 'bms.number_cells', 'operator' : 'lte', 'bound' : 800},
           {'attribute' : 'bms.battery.number_cells', 'operator' : 'gt', 'bound' : 700}]

filter_sort = wf.Filter(filters, 'Filters EBO')

blocks.extend([block_ebo, optimize_ebo, attribute_selection_ebo, filter_sort])
pipes = [wf.Pipe(block_ebo.outputs[0], optimize_ebo.inputs[0]),
         wf.Pipe(optimize_ebo.outputs[1], attribute_selection_ebo.inputs[0]),
         wf.Pipe(attribute_selection_ebo.outputs[0], filter_sort.inputs[0])]

workflow = wf.Workflow(blocks, pipes, filter_sort.outputs[0])
input_values = {0: cells.CELL1_2RC,
                1: limits_voltage_module,
                2: limits_current_module,
                3: limits_voltage_battery,
                4: limits_current_battery,
                5: [33, 34],
                6: [24, 25],
                7: [comb_profile_load,
                    comb_profile_wltp,
                    comb_profile_end],
                12: 5}
#workflow_run = workflow.run(input_values)
#d = workflow_run.to_dict()
#w = wf.WorkflowRun.dict_to_object(d)
#methods_jsonschemas = workflow._method_jsonschemas
#run_default_dict = models.get_jsonschema_default_dict(methods_jsonschemas['run'])
#serialized_run_default_dict = workflow.dict_to_arguments(run_default_dict, 'run')['input_variables_values']

## From serialized default dict to input_values
#sdd_to_iv = {}
#for variable, default_value in serialized_run_default_dict.items():
#    sdd_to_iv[variable] = input_values[variable]
#
#sdd_to_iv['10'] = 2
#run1 = workflow.run(sdd_to_iv)

## Setting values before dict_to_arguments
#run_dict = {}
#for i, input_ in enumerate(block_ebo.inputs):
#    if str(i) in run_default_dict:
#        run_dict[str(i)] = input_values[input_]
#
#run_dict['9'] = 2
#serialized_run_dict = workflow.dict_to_arguments(run_dict, 'run')
#
#run2 = workflow.run(serialized_run_dict['input_variables_values'])
#
## Requetes
#r = c.CreateObject(workflow)
#reference = r.json()
#job = c.SubmitJob(reference['object_class'], reference['id'], 'run', run_dict)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Nov 24 19:24:53 2021

@author: steven
"""

import math
import dessia_common as dc
from dessia_common.utils.types import isinstance_base_types, is_sequence, full_classname

# def basic_dict_diff(dict1, dict2):
#     missing_keys_in_other_object = []
#     diff_values = []
    
#     for key, value in dict1.items():
#         if key not in dict2:
#             missing_keys_in_other_object.append(key)
#         else:                
#             other_value = dict2[key]
#             if value != other_value:
#                 diff_values.append(key, value, other_value)
            
#     return diff_values, missing_keys_in_other_object


def diff(value1, value2, path='#'):
    diff_values = []
    missing_keys_in_other_object = []
    invalid_types = []
    
    if type(value1) != type(value2):
        invalid_types.append(path)
        return diff_values, missing_keys_in_other_object, invalid_types
    
    if isinstance_base_types(value1):
        if isinstance(value1, float) and \
                math.isclose(value1, value2, abs_tol=dc.FLOAT_TOLERANCE):
            return diff_values, missing_keys_in_other_object, invalid_types
        elif value1 != value2:
            diff_values.append((path, value1, value2))
        return diff_values, missing_keys_in_other_object, invalid_types
    elif is_sequence(value1):
        return sequence_diff(value1, value2, path=path)
    elif isinstance(value1, dict):
        return dict_diff(value1, value2, path=path)
    # elif hasattr(value1, '_data_eq'):
    else:
        # Should be object
        if hasattr(value1, '_data_eq'):
            # DessiaObject
            if value1._data_eq(value2):
                return [], [], []
        if value1 == value2:
            return [], [], []
        
        # 
        # return diff_values, missing_keys_in_other_object, invalid_types
        raise NotImplementedError('Undefined type in diff: {}'.format(type(value1)))


def dict_diff(dict1, dict2, path='#'):
    missing_keys_in_other_object = []
    diff_values = []
    invalid_types = []
    
    for key, value in dict1.items():
        path_key = '{}/{}'.format(path, key)
        if key not in dict2:
            missing_keys_in_other_object.append(key)
        else:                
            dk, mkk, itk = diff(value, dict2[key], path=path_key)
            diff_values.extend(dk)
            missing_keys_in_other_object.extend(mkk)
            invalid_types.extend(itk)
            
    return diff_values, missing_keys_in_other_object, invalid_types

    
def sequence_diff(seq1, seq2, path='#'):
    diff_values = []
    missing_keys_in_other_object = []
    invalid_types = []
    
    if len(seq1) != len(seq2):
        diff_values.append((path, seq1, seq2))
    else:
        for iv, (v1, v2) in enumerate(zip(seq1, seq2)):
            path_value = '{}/{}'.format(path, iv)
            dv, mkv, itv = diff(v1, v2, path=path_value)
            # print('dvs', dv, v1, v2)
            diff_values.extend(dv)
            missing_keys_in_other_object.extend(mkv)
            invalid_types.extend(itv)
    return diff_values, missing_keys_in_other_object, invalid_types


def data_eq(value1, value2):
    if not isinstance(value2, type(value1))\
            and not isinstance(value1, type(value2)):
        return False
    
    if isinstance_base_types(value1):
        if isinstance(value1, float):
            return math.isclose(value1, value2, abs_tol=dc.FLOAT_TOLERANCE)
        else:
            # if not(value1 == value2):
                # print('base types', value1, value2, value1 == value2)
            return value1 == value2
        
    if isinstance(value1, dict):
        return dict_data_eq(value1, value2)

    if is_sequence(value1):
        return sequence_data_eq(value1, value2)
        
    # Else: its an object

    if full_classname(value1) != full_classname(value2):
        # print('full classname !=')
        return False

    # Test if _data_eq is customized
    if hasattr(value1, '_data_eq'):
        custom_method = (value1._data_eq.__code__
                                 is not dc.DessiaObject._data_eq.__code__)
        if custom_method:
            return value1._data_eq(value2)

    # Not custom, use generic implementation
    eq_dict = value1._serializable_dict()
    if 'name' in eq_dict:
        del eq_dict['name']
        
    other_eq_dict = value2._serializable_dict()

    return dict_data_eq(eq_dict, other_eq_dict)


def dict_data_eq(dict1, dict2):
    
    for key, value in dict1.items():
        if key not in dict2:
            # print('missing key', key)
            return False
        else:                
            if not data_eq(value, dict2[key]):
                # print('dict key !=', key, value, dict2[key])
                return False
    return True
                            
    
def sequence_data_eq(seq1, seq2):
    if len(seq1) != len(seq2):
        # print('len diff', seq1, seq2)
        return False
    else:
        for iv, (v1, v2) in enumerate(zip(seq1, seq2)):
            if not data_eq(v1, v2):
                # print('seq false')
                return False
        
    return True


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""

from dessia_common import DessiaObject
import dessia_common.typings as dct
from typing import List, Dict


class Submodel(DessiaObject):
    _generic_eq = True

    def __init__(self, subvalue: int, name: str = ''):
        self.subvalue = subvalue
        self.name = name

        DessiaObject.__init__(self, name=name)


class Model(DessiaObject):
    _generic_eq = True

    def __init__(self, value: int, submodel: Submodel, name: str = ''):
        self.value = value
        self.submodel = submodel

        DessiaObject.__init__(self, name=name)


class Generator(DessiaObject):
    def __init__(self, parameter: int, nb_solutions: int = 25, name: str = ''):
        self.parameter = parameter
        self.nb_solutions = nb_solutions
        self.models = None

        DessiaObject.__init__(self, name=name)

    def generate(self) -> None:
        submodels = [Submodel(self.parameter * i)
                     for i in range(self.nb_solutions)]
        self.models = [Model(self.parameter + i, submodels[i])
                       for i in range(self.nb_solutions)]


class Optimizer(DessiaObject):
    def __init__(self, model_to_optimize: Model, name: str = ''):
        self.model_to_optimize = model_to_optimize

        DessiaObject.__init__(self, name=name)

    def optimize(self, optimization_value: int = 3) -> None:
        self.model_to_optimize.value += optimization_value


class Component(DessiaObject):
    def __init__(self, efficiency, name: str = ''):
        self.efficiency = efficiency
        DessiaObject.__init__(self, name=name)

    def power_simulation(self, power_value: dct.Power):
        return power_value * self.efficiency


class ComponentConnection(DessiaObject):
    def __init__(self, input_component: Component,
                 output_component: Component, name: str = ''):
        self.input_component = input_component
        self.output_component = output_component
        DessiaObject.__init__(self, name=name)


class SystemUsage(DessiaObject):
    def __init__(self, time: List[dct.Time], power: List[dct.Power],
                 name: str = ''):
        self.time = time
        self.power = power
        DessiaObject.__init__(self, name=name)


class System(DessiaObject):
    _dessia_methods = ['power_simulation']

    def __init__(self, components: List[Component],
                 component_connections: List[ComponentConnection],
                 name: str = ''):
        self.components = components
        self.component_connections = component_connections
        DessiaObject.__init__(self, name=name)

    def output_power(self, input_power: dct.Power):
        return input_power * 0.8

    def power_simulation(self, usage: SystemUsage):
        output_power = []
        for time, input_power in zip(usage.time, usage.power):
            output_power.append(self.output_power(input_power))
        return SystemSimulationResult(self, usage, output_power)


class SystemSimulationResult(DessiaObject):
    def __init__(self, system: System, system_usage: SystemUsage,
                 output_power: List[dct.Power], name: str = ''):
        self.system = system
        self.system_usage = system_usage
        self.output_power = output_power
        DessiaObject.__init__(self, name=name)


class SystemSimulationList(DessiaObject):
    def __init__(self, simulations: List[SystemSimulationResult],
                 name: str = ''):
        self.simulations = simulations
        DessiaObject.__init__(self, name=name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A simple workflow composed of functions
"""

import math
import dessia_common.workflow as workflow

def sinus_f(x: float) -> float:
    return math.sin(x)

def asin_f(x: float) -> float:
    return math.asin(x)


sinus = workflow.Function(sinus_f)
arcsin = workflow.Function(asin_f)

pipe1 = workflow.Pipe(sinus.outputs[0], arcsin.inputs[0])

workflow = workflow.WorkFlow([sinus, arcsin], [pipe1])

workflow.plot_graph()

workflow_run = workflow.run([math.pi/3])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Script to test abilities of DessiaObject class
"""

import dessia_common

class Model(dessia_common.DessiaObject):
    def __init__(self, a, b, name=''):
        dessia_common.DessiaObject.__init__(self, a=a, b=b, name=name)
        
        
model = Model(2, 3.4, 'name of model')

print(model.to_dict())


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 16 12:32:31 2021

@author: steven
"""

from typing import List

def istep_from_value_on_list(list_:List[float], value:float,
                             extrapolate=False):
    for ipoint, (point1, point2) in enumerate(zip(list_[:-1],
                                                  list_[1:])):

        point1_s, point2_s = sorted((point1, point2))
        if (point1_s <= value) and (value <= point2_s):
            alpha = (value-point1_s)/(point2_s-point1_s)
            if alpha < 0 or alpha > 1:
                raise ValueError
            return ipoint + alpha
    # values = [p for p in list_]


    if extrapolate:
        if abs(list_[0]-value) < abs(list_[-1]-value):
            # Closer to left
            return 0.
        else:
            # Closer to right
            return len(list_)-1


    min_values = min(list_)
    max_values = max(list_)
    raise ValueError('Specified value not found in list_: {} not in [{}, {}]'.format(value, min_values, max_values))


def interpolate_from_istep(objects, istep:float):
    n_objects = len(objects)
    if (istep < 0) or (istep > n_objects - 1):
        raise(ValueError('Extrapolating is not supported'))
    istep1 = int(istep)
    if istep1 == istep:
        # No interpolation needed
        return objects[int(istep)]
    else:
        alpha = istep - istep1
        point1 = objects[istep1]
        point2 = objects[istep1+1]
        return (1-alpha)*point1+(alpha)*point2


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""

from dessia_common.forms import StandaloneSubobject, StandaloneObject,\
    EmbeddedSubobject, EnhancedStandaloneSubobject,\
    InheritingStandaloneSubobject

standalone_subobject = StandaloneSubobject(floatarg=2.3)
embedded_subobject = EmbeddedSubobject(name="ES0")
dynamic_dict = {'a': True, 'b': False, 'c': False}
static_dict = {'name': 'Static Dict',
               'float_value': 93.89,
               'int_value': 10,
               'is_valid': True}

tuple_arg = ('tuple', 120)
sublist_ss = StandaloneSubobject(floatarg=-568.1)
sublist_es = EmbeddedSubobject()
enhanced_ss = EnhancedStandaloneSubobject(floatarg=-568.1, boolarg=False)
inheriting_list = [EnhancedStandaloneSubobject(floatarg=-7516.15,
                                               boolarg=True),
                   InheritingStandaloneSubobject(floatarg=1561.57,
                                                 strarg='Subclass')]

standalone_object = StandaloneObject(standalone_subobject=standalone_subobject,
                                     embedded_subobject=embedded_subobject,
                                     dynamic_dict=dynamic_dict,
                                     tuple_arg=tuple_arg,
                                     intarg=1, strarg='test_str',
                                     object_list=[sublist_ss],
                                     subobject_list=[sublist_es],
                                     builtin_list=[0, 1, 2, 3],
                                     union_arg=enhanced_ss,
                                     subclass_arg=inheriting_list[0],
                                     array_arg=[[0, 1], [1, 0]])





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb 17 17:00:35 2020

@author: jezequel
"""

import pkg_resources

from dessia_common.vectored_objects import Catalog, Objective, ParetoSettings,\
    ObjectiveSettings, pareto_frontier
import os

choice_args = ['MPG', 'Cylinders', 'Displacement', 'Horsepower',
               'Weight', 'Acceleration', 'Model']  # Ordered

minimized_attributes = {'MPG': False, 'Horsepower': True,
                        'Weight': True, 'Acceleration': False}
coefficients = {'Cylinders': 0, 'MPG': -0.70,  'Displacement': 0,
                'Horsepower': 0, 'Weight': 0.70, 'Acceleration': 0, 'Model': 0}

pareto_settings = ParetoSettings(minimized_attributes=minimized_attributes,
                                 enabled=True)

# objective_settings0 = ObjectiveSettings(n_near_values=4, enabled=True)
# objective0 = Objective(coefficients=coefficients,
#                         settings=objective_settings0,
#                         name='ObjectiveName')

# dirname = os.path.dirname(__file__)
# relative_filepath = './data/cars.csv'
# filename = os.path.join(dirname, relative_filepath)

csv_cars = pkg_resources.resource_stream('dessia_common', 'models/data/cars.csv')

# array, variables = from_csv(filename=csv_cars, end=None,
#                             remove_duplicates=True)

catalog = Catalog.from_csv(csv_cars)
catalog.name = 'Cars dataset'
catalog.pareto_settings = pareto_settings

# catalog.plot()

# from plot_data import plot_canvas

# plot_canvas(plot_data_object=catalog.plot_data()[0], canvas_id='canvas')
# cost = catalog.build_costs(pareto_settings)
# print('cost', cost)

# p_f = pareto_frontier(cost)
# print('p_f', p_f)



from string import Template

workflow_template = Template('''
<!DOCTYPE html>
<html>
<head>
	<title>Dessia Workflow</title>
    <link rel="stylesheet" type="text/css" href="https://cdnjs.cloudflare.com/ajax/libs/jointjs/2.1.0/joint.css" />
</head>
<body>
    <!-- content -->
    <div id="myholder" ></div>

    <!-- dependencies -->
    <script src="https://cdnjs.cloudflare.com/ajax/libs/jquery/3.4.1/jquery.js"></script>
    <script src="https://cdnjs.cloudflare.com/ajax/libs/lodash.js/3.10.1/lodash.js"></script>
    <script src="https://cdnjs.cloudflare.com/ajax/libs/backbone.js/1.4.0/backbone.js"></script>
    <script src="https://cdnjs.cloudflare.com/ajax/libs/jointjs/3.1.1/joint.js"></script>

    <!-- code -->
    <script type="text/javascript">
            var workflow_data = $workflow_data;
			var blocks_data = workflow_data['blocks'];
			var edges_data = workflow_data['edges'];
			var nonblock_variables_data = workflow_data['nonblock_variables'];


			var graph = new joint.dia.Graph;
			// new joint.dia.Paper({ el: $$('#paper-create'), width: 650, height: 200, gridSize: 1, model: graph });
			var paper = new joint.dia.Paper({
          el: document.getElementById('myholder'),
          model: graph,
          width: '100%',
          height: 900,
          gridSize: 25,
					drawGrid: true,
      });


	    paper.options.defaultConnector = {
	      name: 'jumpover',
	      args: {
	        radius: 5
	      }
	    };

	    paper.options.defaultRouter = {
	      name: 'manhattan',
	      args: {
	        padding: 25
	      }
	    };

	    joint.dia.Link.define('standard.Link', {
	      attrs: {
	        line: {
	          connection: true,
	          stroke: '#333',
	          strokeWidth: 2,
	          strokeLinejoin: 'round',
	          targetMarker: {
	            'type': 'path',
	            'd': 'M 10 -5 0 0 10 5 z'
	            }
	          },
	          wrapper: {
	            connection: true,
	            strokeWidth: 10,
	            strokeLinejoin: 'round'
	            }
	          }
	      }, {
	      markup: [{
	        tagName: 'path',
	        selector: 'wrapper',
	        attributes: {
	          'fill': 'none',
	          'cursor': 'pointer',
	          'stroke': 'transparent'
	          }
	        }, {
	        tagName: 'path',
	        selector: 'line',
	        attributes: {
	          'fill': 'none',
	          'pointer-events': 'none'
	          }
	        }]
	      });


			var blocks = [];
			for (const [i, block_data] of blocks_data.entries()){

				var block = new joint.shapes.devs.Model({
						position: { x: block_data['position'][0],
						 						y: block_data['position'][1]},
						size: { width: 220, height: 120 },
						ports: {
								groups: {
										'in': {
											label: {
					                    position: {
					                        name: 'right', args: {
					                            x: 12,
																			y: 5,
					                        }}},
												attrs: {
														'.port-label': {fontSize: 8},
														'.port-body': {
															fill: '#DDD',
															stroke: 'black',
															strokeWidth: 2,
															height: 10,
															width: 10,
															magnet: 'passive'
														}
												},
												markup: '<rect class="port-body"/>'
										},
										'out': {
											label: {
					                    // label layout definition:
					                    position: {
					                        name: 'left', args: {
					                            x: -12,
																			y: 5,
					                        }}},
												attrs: {
														'.port-label': {fontSize: 8},
														'.port-body': {
																fill: '#CCC',
																stroke: 'black',
																strokeWidth: 2,
																height: 10,
																width: 10,
																magnet: true,
																'ref-x': -10
														}
												},
												markup: '<rect class="port-body"/>'
										}
								}
						},
						attrs: {
								'.label': { text: block_data["name"],
								 					'ref-x': .5, 'ref-y': .05,
													'font-size': 12 },
								rect: { fill: '#EEE' }
						}
				});

				// var input_ports = [];
				let j = 0;
				for (port_input_data of block_data['inputs']){
					var port = {
						id: String(i)+String(j),
		        group: 'in',
		        args: {},
		        attrs: { text: { text: port_input_data['name'] } },
		    	};
					block.addPort(port);
					if (port_input_data['is_workflow_input']){
						if (port_input_data['has_default_value']){
							block.portProp(port, 'attrs/.port-body/fill', '#FFA500');
						}
						else {
							block.portProp(port, 'attrs/.port-body/fill', '#00FF00');
						}
					}

					j++;
				}

				for (port_output_data of block_data['outputs']){
					var port = {
						id: String(i)+String(j),
		        group: 'out',
		        args: {},
		        attrs: { text: { text: port_output_data['name'] } },
		    	};
					block.addPort(port);
					if (port_output_data['is_workflow_output']){
						block.portProp(port, 'attrs/.port-body/fill', '#ff1919');
					}
					j++;
				}


			graph.addCell(block);
			blocks.push(block);
		}

		var nonblock_variables = []
		for (nonblock_variable_data of nonblock_variables_data){
			var variable = new joint.shapes.devs.Model({
					position: {x: nonblock_variable_data['position'][0],
						 				 y: nonblock_variable_data['position'][1]},
					size: { width: 100, height: 20 },
					attrs: {
							'.label': { text: nonblock_variable_data["name"],
												'ref-x': .5, 'ref-y': .05,
												'font-size': 12 },
							rect: { fill: '#EEE' }
					}
			});


		graph.addCell(variable);
		nonblock_variables.push(variable);
		}


		for (edge of edges_data){
			var node1 = edge[0];
			if (typeof(node1) == 'number'){
				var source = {'id': nonblock_variables[node1]};
			}
			else{
				var block1 = blocks[node1[0]];
				var port1 = block1.getPorts()[node1[1]];
				var source = {
					id: block1.id,
					port: port1.id
				}
			}

			var node2 = edge[1];
			if (typeof(node2) == 'number'){
				var target = {'id': nonblock_variables[node2]};
			}
			else{
				var block2 = blocks[node2[0]];
				var port2 = block2.getPorts()[edge[1][1]]
				var target = {
					id: block2.id,
					port: port2.id
				}
			}

			var link = new joint.shapes.standard.Link({
			      source: source,
			      target: target
			    });
			// Assume graph has the srcModel and dstModel with in and out ports.
			graph.addCell(link)
		}



    </script>
</body>
</html>
''')


visjs_template = Template('''
<html>
        <head>
        <script src="https://cdnjs.cloudflare.com/ajax/libs/vis/4.21.0/vis.min.js"></script>
        <link href="https://cdnjs.cloudflare.com/ajax/libs/vis/4.21.0/vis.min.css" rel="stylesheet" />
        <meta charset="utf-8"/>
        <style type="text/css">
            #mynetwork {
                border: 1px solid lightgray;
            }
        </style>
        <title>$name</title>
    </head>
    <body>
    <div id="mynetwork"></div>

    <script type="text/javascript">
    var nodes = new vis.DataSet($nodes);

var edges = new vis.DataSet($edges);


// create a network
    var container = document.getElementById('mynetwork');

    // provide the data in the vis format
    var data = {
        nodes: nodes,
        edges: edges
    };

    scale=function (min,max,total,value) {
      if (max === min) {
        return 0.5;
      }
      else {
        var scale = 1 / (max);
        return Math.max(0,(value )*scale);
      }
    }

    var options = {edges: {scaling: {'customScalingFunction': scale}}};

    // initialize your network!
    var network = new vis.Network(container, data, options);
</script>
</body>
</html>
''')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 19 17:25:40 2021

@author: masfaraud
"""
import dessia_common.workflow as dcw
import dessia_common.typings as dct
import dessia_common.tests as dctest


instanciate_system = dcw.InstantiateModel(model_class=dctest.System,
                                          name='Instantiate Generator')

instanciate_system.outputs[0].memorize = True

simulate = dcw.ModelMethod(method_type=dct.MethodType(dctest.System, 'power_simulation'),
                           name='Generator Generate')

pipe_1 = dcw.Pipe(input_variable=instanciate_system.outputs[0],
                  output_variable=simulate.inputs[0])


blocks = [instanciate_system, simulate]
pipes = [pipe_1]
simulation_workflow = dcw.Workflow(blocks=blocks, pipes=pipes,
                                   output=simulate.outputs[0])




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 19 17:19:08 2021

@author: masfaraud
"""

import dessia_common.tests

components1 = [dessia_common.tests.Component(0.7),
              dessia_common.tests.Component(0.9),
              dessia_common.tests.Component(0.95)]

components2 = [dessia_common.tests.Component(0.7),
              dessia_common.tests.Component(0.96),
              dessia_common.tests.Component(0.95),
              dessia_common.tests.Component(0.99)]


component_connections1 = [dessia_common.tests.ComponentConnection(components1[0], components1[1]),
                         dessia_common.tests.ComponentConnection(components1[1], components1[2])
                         ]

component_connections2 = [dessia_common.tests.ComponentConnection(components2[0], components2[1]),
                          dessia_common.tests.ComponentConnection(components2[1], components2[2]),
                          dessia_common.tests.ComponentConnection(components2[2], components2[3])
                         ]


system1 = dessia_common.tests.System(components1, component_connections1, 'system1')#{components1[1]: 0.32})
system2 = dessia_common.tests.System(components2, component_connections2, 'system2')# {components2[0]: 0.32})

usage1 = dessia_common.tests.SystemUsage([0, 1., 2., 5., 8.], [0., 100., 150, 350, 120.])
usage2 = dessia_common.tests.SystemUsage([0, 1., 2., 5., 10.], [0., 76., 150, 357, 110.])


simulation1 = system1.power_simulation(usage1)
simulation2 = system2.power_simulation(usage1)
simulation3 = system1.power_simulation(usage2)
simulation4 = system2.power_simulation(usage2)


simulation_list = dessia_common.tests.SystemSimulationList([simulation1, simulation2, simulation3, simulation4])



from dessia_common.workflow import InstantiateModel, ModelMethod,\
    TypedVariable, TypedVariableWithDefaultValue, ModelAttribute,\
    Pipe, Workflow, WorkflowBlock, ForEach, MultiPlot
from dessia_common.forms import Generator, Optimizer
from dessia_common import MethodType

instanciate_generator = InstantiateModel(model_class=Generator,
                                         name='Instantiate Generator')

generate_method = MethodType(class_=Generator, name='generate')
generator_generate = ModelMethod(method_type=generate_method,
                                 name='Generator Generate')
attribute_selection = ModelAttribute(attribute_name='models',
                                     name='Attribute Selection')

# Subworkflow of model optimization
instanciate_optimizer = InstantiateModel(model_class=Optimizer,
                                         name='Instantiate Optimizer')


generate_method = MethodType(class_=Optimizer, name='optimize')
optimization = ModelMethod(method_type=generate_method, name='Optimization')

model_fetcher = ModelAttribute(attribute_name='model_to_optimize',
                               name='Model Fetcher')

pipe1_opt = Pipe(input_variable=instanciate_optimizer.outputs[0],
                 output_variable=optimization.inputs[0])
pipe2_opt = Pipe(input_variable=optimization.outputs[1],
                 output_variable=model_fetcher.inputs[0])
optimization_blocks = [instanciate_optimizer, optimization, model_fetcher]
optimization_pipes = [pipe1_opt, pipe2_opt]
optimization_workflow = Workflow(blocks=optimization_blocks,
                                 pipes=optimization_pipes,
                                 output=model_fetcher.outputs[0],
                                 name='Optimization Workflow')

optimization_workflow_block = WorkflowBlock(workflow=optimization_workflow,
                                            name='Workflow Block')

parallel_optimization = ForEach(workflow_block=optimization_workflow_block,
                                iter_input_index=0, name='ForEach')

display_attributes = ['intarg', 'strarg', 'standalone_subobject/floatarg']
display = MultiPlot(attributes=display_attributes, name='Display')

int_variable = TypedVariable(type_=int, name="Some Integer")
name_variable = TypedVariableWithDefaultValue(type_=str, name="Shared Name",
                                              default_value="Shared Name")

pipe_int_1 = Pipe(input_variable=int_variable,
                  output_variable=instanciate_generator.inputs[1])
pipe_name_1 = Pipe(input_variable=name_variable,
                   output_variable=instanciate_generator.inputs[2])
pipe_name_2 = Pipe(input_variable=name_variable,
                   output_variable=parallel_optimization.inputs[1])
pipe_1 = Pipe(input_variable=instanciate_generator.outputs[0],
              output_variable=generator_generate.inputs[0])
pipe_2 = Pipe(input_variable=generator_generate.outputs[1],
              output_variable=attribute_selection.inputs[0])
pipe_3 = Pipe(input_variable=attribute_selection.outputs[0],
              output_variable=parallel_optimization.inputs[0])
pipe_4 = Pipe(input_variable=parallel_optimization.outputs[0],
              output_variable=display.inputs[0])

blocks = [instanciate_generator, generator_generate,
          attribute_selection, parallel_optimization, display]
pipes = [pipe_int_1, pipe_name_1, pipe_name_2, pipe_1, pipe_2, pipe_3, pipe_4]
workflow_ = Workflow(blocks=blocks, pipes=pipes,
                     output=parallel_optimization.outputs[0])


from typing import TypeVar, Generic, Dict, Any, Tuple


T = TypeVar('T')


class Subclass(Generic[T]):
    pass


class InstanceOf(Generic[T]):
    pass


class MethodType(Generic[T]):
    def __init__(self, class_: T, name: str):
        self.class_ = class_
        self.name = name

    def __deepcopy__(self):
        return MethodType(self.class_, self.name)


class ClassMethodType(MethodType[T]):
    def __init__(self, class_: T, name: str):
        MethodType.__init__(self, class_=class_, name=name)


class AttributeType(Generic[T]):
    def __init__(self, class_: T, name: str):
        self.class_ = class_
        self.name = name


class ClassAttributeType(AttributeType[T]):
    def __init__(self, class_: T, name: str):
        AttributeType.__init__(self, class_=class_, name=name)


# Types Aliases
JsonSerializable = Dict[str, Any]
RGBColor = Tuple[float, float, float]


# Measures
class Measure(float):
    units = ''


class Distance(Measure):
    units = 'm'


class Torque(Measure):
    units = 'Nm'


class Stress(Measure):
    units = 'Pa'


class Time(Measure):
    units = 's'


class Speed(Measure):
    units = 'm/s'


class Acceleration(Measure):
    units = 'm/s²'


class Mass(Measure):
    units = 'Kg'


class Force(Measure):
    units = 'N'


class Work(Measure):
    units = 'N*m'


class Power(Measure):
    units = 'N*m/s'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""

import io


class BinaryFile(io.BytesIO):
    def __init__(self, filename: str = ''):
        super().__init__()
        self.filename = filename


class StringFile(io.StringIO):
    def __init__(self, filename: str = ''):
        super().__init__()
        self.filename = filename


class XLSXFile(BinaryFile):
    extension = 'xlsx'


class XLSFile(BinaryFile):
    extension = 'xls'


class TextFile(StringFile):
    extension = 'txt'


class CSVFile(StringFile):
    extension = 'csv'


class MarkdownFile(StringFile):
    extension = 'md'



# -*- coding: utf-8 -*-
"""

"""

from setuptools import setup

from os.path import dirname, isdir, join
import re
from subprocess import CalledProcessError, check_output


def readme():
    with open("README.md") as f:
        return f.read()


tag_re = re.compile(r"\btag: %s([0-9][^,]*)\b")
version_re = re.compile("^Version: (.+)$", re.M)


def version_from_git_describe(version):
    if version[0] == "v":
        version = version[1:]
    # PEP 440 compatibility
    number_commits_ahead = 0
    if "-" in version:
        version, number_commits_ahead, commit_hash = version.split("-")
        number_commits_ahead = int(number_commits_ahead)
    # print('number_commits_ahead', number_commits_ahead)

    split_versions = version.split(".")
    if "post" in split_versions[-1]:
        suffix = split_versions[-1]
        split_versions = split_versions[:-1]
    else:
        suffix = None
    for pre_release_segment in ["a", "b", "rc"]:
        if pre_release_segment in split_versions[-1]:
            if number_commits_ahead > 0:
                split_versions[-1] = str(
                    split_versions[-1].split(pre_release_segment)[0]
                )
                if len(split_versions) == 2:
                    split_versions.append("0")
                if len(split_versions) == 1:
                    split_versions.extend(["0", "0"])
                split_versions[-1] = str(int(split_versions[-1]) + 1)
                future_version = ".".join(split_versions)
                return "{}.dev{}".format(future_version, number_commits_ahead)
            else:
                return ".".join(split_versions)
    if number_commits_ahead > 0:
        if len(split_versions) == 2:
            split_versions.append("0")
        if len(split_versions) == 1:
            split_versions.extend(["0", "0"])
        split_versions[-1] = str(int(split_versions[-1]) + 1)
        split_versions = ".".join(split_versions)
        return "{}.dev{}".format(split_versions, number_commits_ahead)
    else:
        if suffix is not None:
            split_versions.append(suffix)
        return ".".join(split_versions)


# Just testing if get_version works well
assert version_from_git_describe("v0.1.7.post2") == "0.1.7.post2"
assert version_from_git_describe("v0.0.1-25-gaf0bf53") == "0.0.2.dev25"
assert version_from_git_describe("v0.1-15-zsdgaz") == "0.1.1.dev15"
assert version_from_git_describe("v1") == "1"
assert version_from_git_describe("v1-3-aqsfjbo") == "1.0.1.dev3"


def get_version():
    # Return the version if it has been injected into the file by git-archive
    version = tag_re.search("$Format:%D$")
    if version:
        return version.group(1)
    d = dirname(__file__)

    if isdir(join(d, ".git")):
        cmd = "git describe --tags"
        try:
            version = check_output(cmd.split()).decode().strip()[:]
        except CalledProcessError:
            raise RuntimeError("Unable to get version number from git tags")
        return version_from_git_describe(version)
    else:
        # Extract the version from the PKG-INFO file.
        with open(join(d, "PKG-INFO")) as f:
            version = version_re.search(f.read()).group(1)
    return version


setup(
    name="dessia_common",
    version=get_version(),
    description="Common tools for DessIA software",
    long_description=readme(),
    keywords=["Dessia", "SDK", "engineering"],
    url="https://github.com/Dessia-tech/dessia-common",
    author="Steven Masfaraud",
    author_email="masfaraud@dessia.tech",
    include_package_data=True,
    packages=[
        "dessia_common",
        "dessia_common.utils",
        "dessia_common.models",
        "dessia_common.models.workflows",
    ],
    install_requires=[
        "typeguard",
        "networkx",
        "numpy",
        "pandas",
        "mypy_extensions",
        "scipy",
        "pyDOE",
        "dectree",
        "openpyxl",
        "parameterized",
    ],
    python_requires=">=3.7",
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""

import sys
import collections
import numpy as npy

import dessia_common.core
from dessia_common.utils.types import is_sequence


def get_in_object_from_path(object_, path):
    segments = path.lstrip('#/').split('/')
    if isinstance(object_, dict):
        try:
            element = object_[segments[0]]
        except KeyError:
            print(object_, segments[0])

            raise RuntimeError('Cannot get in {} {}: end up @ {}'.format(object_, path, segments[0]))
    else:
        element = getattr(object_, segments[0])

    for segment in segments[1:]:
        if is_sequence(element):
            element = element[int(segment)]
        elif isinstance(element, dict):  # A dict?
            if segment in element:
                element = element[segment]
            else:
                element = element[int(segment)]
        else:
            element = getattr(element, segment)

    return element


def merge_breakdown_dicts(dict1, dict2):
    dict3 = dict1.copy()
    for class_name, refs in dict2.items():
        if class_name in dict3:
            # Decide by lower depth
            for obj, path in refs.items():
                if obj in dict3[class_name]:
                    if len(path.split('.')) < len(dict3[class_name][obj].split('.')):
                        dict3[class_name][obj] = path
                else:
                    dict3[class_name][obj] = path
            # dict3[class_name].update(refs)
        else:
            dict3[class_name] = refs
    return dict3


def object_breakdown(obj, path=''):
    bd_dict = {}
    if obj is None:
        return bd_dict

    if (isinstance(obj, str) or isinstance(obj, float) or isinstance(obj, int)):
        return bd_dict

    if isinstance(obj, npy.ndarray):
        return bd_dict

    if isinstance(obj, list) or isinstance(obj, tuple) or isinstance(obj, set):
        if path:
            path += '.'

        for i, li in enumerate(obj):
            path2 = path + '{}'.format(i)
            bd_dict = merge_breakdown_dicts(bd_dict, object_breakdown(li, path2))
    elif isinstance(obj, dict):
        if path:
            path += '.'

        for k, v in obj.items():
            path2 = path + str(k)
            bd_dict = merge_breakdown_dicts(bd_dict, object_breakdown(v, path2))
    else:
        # Put object and break it down
        if path:  # avoid to get root object
            if hasattr(obj, '__dict__'):
                if obj.__class__.__name__ in bd_dict:
                    if obj in bd_dict[obj.__class__.__name__]:
                        if len(path.split('.')) < len(bd_dict[obj.__class__.__name__][obj].split('.')):
                            bd_dict[obj.__class__.__name__][obj] = path
                    else:
                        bd_dict[obj.__class__.__name__][obj] = path
                else:
                    bd_dict[obj.__class__.__name__] = collections.OrderedDict()
                    bd_dict[obj.__class__.__name__][obj] = path

        if path:
            path += '.'

        if isinstance(obj, dessia_common.core.DessiaObject):
            obj_dict = obj._serializable_dict()
        else:
            if hasattr(obj, '__dict__'):
                obj_dict = obj.__dict__
            else:
                obj_dict = {}

        for k, v in obj_dict.items():
            # dict after lists
            if not (isinstance(v, dict)
                    or isinstance(v, list)
                    or isinstance(v, tuple)
                    ):  # Should be object or builtins
                dict2 = object_breakdown(v, path=path + k)
                bd_dict = merge_breakdown_dicts(bd_dict, dict2)

        for k, v in obj_dict.items():
            # First lists and tuples
            if isinstance(v, list) or isinstance(v, tuple):
                dict2 = object_breakdown(v, path=path + k)
                bd_dict = merge_breakdown_dicts(bd_dict, dict2)

        for k, v in obj_dict.items():
            # dict after lists
            if isinstance(v, dict):
                dict2 = object_breakdown(v, path=path + k)
                bd_dict = merge_breakdown_dicts(bd_dict, dict2)

    return bd_dict


def deep_getsizeof(o, ids=None):
    """Find the memory footprint of a Python object

    This is a recursive function that drills down a Python object graph
    like a dictionary holding nested dictionaries with lists of lists
    and tuples and sets.

    The sys.getsizeof function does a shallow size of only. It counts each
    object inside a container as pointer only regardless of how big it
    really is.

    :param o: the object
    :param ids:
    :return:
    """

    if ids is None:
        ids = set()

    d = deep_getsizeof
    if id(o) in ids:
        return 0

    r = sys.getsizeof(o)
    ids.add(id(o))

    if isinstance(o, str):
        return r

    # if isinstance(o, collections.Mapping):
    #     return r + sum(d(k, ids) + d(v, ids) for k, v in o.items())

    if isinstance(o, collections.Mapping):
        return r + sum(d(k, ids) + d(v, ids) for k, v in o.items())

    if isinstance(o, collections.Container):
        return r + sum(d(x, ids) for x in o.__dict__.values())

    return r


def breakdown_analysis(obj):
    obj_breakdown = object_breakdown(obj)

    stats = {}
    stats['total_size'] = deep_getsizeof(obj)

    number_objs = {}
    size_objs = {}
    meansize_objs = {}
    for class_name, class_objs in obj_breakdown.items():
        nobjs = len(class_objs)
        number_objs[class_name] = nobjs
        size_objs[class_name] = deep_getsizeof(class_objs)
        meansize_objs[class_name] = size_objs[class_name] / nobjs
    stats['subobjects_number_by_class'] = number_objs
    stats['subobjects_size_by_class'] = size_objs
    stats['subobjects_meansize_by_class'] = meansize_objs
    return stats



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""
import io
import sys
import warnings
import math
import random
import copy
from functools import reduce
import collections
from copy import deepcopy
import inspect
import json
from dessia_common.exports import XLSXWriter

import dessia_common.errors
from dessia_common.utils.diff import data_eq, diff
from dessia_common.utils.serialization import dict_to_object, serialize_dict_with_pointers
from dessia_common.utils.types import is_jsonable, is_builtin, get_python_class_from_class_name, serialize_typing, full_classname, is_sequence, isinstance_base_types, is_typing, TYPING_EQUIVALENCES
from dessia_common.utils.copy import deepcopy_value

from typing import List, Dict, Type, Tuple, Union, Any, TextIO, BinaryIO, \
    get_type_hints, get_origin, get_args
try:
    from typing import TypedDict  # >=3.8
except ImportError:
    from mypy_extensions import TypedDict  # <=3.7
import traceback as tb
from dessia_common.typings import Measure, JsonSerializable,\
    Subclass, InstanceOf, MethodType, ClassMethodType

from importlib import import_module


JSONSCHEMA_HEADER = {"definitions": {},
                     "$schema": "http://json-schema.org/draft-07/schema#",
                     "type": "object",
                     "required": [],
                     "properties": {}}


TYPES_STRINGS = {int: 'int', float: 'float', bool: 'boolean', str: 'str',
                 list: 'list', tuple: 'tuple', dict: 'dict'}

SEQUENCE_TYPINGS = ['List', 'Sequence', 'Iterable']

_FORBIDDEN_ARGNAMES = ['self', 'cls', 'progress_callback', 'return']

TYPES_FROM_STRING = {'unicode': str, 'str': str, 'float': float,
                     'int': int, 'bool': bool}


# DEPRECATED_ATTRIBUTES = {'_editable_variss' : '_allowed_methods'}
def deprecated(use_instead=None):
    def decorated(function):
        def wrapper(*args, **kwargs):
            deprecation_warning(function.__name__, 'Function', use_instead)
            print('Traceback : ')
            tb.print_stack(limit=2)
            return function(*args, **kwargs)

        return wrapper

    return decorated


def deprecation_warning(name, object_type, use_instead=None):
    warnings.simplefilter('once', DeprecationWarning)
    msg = "\n\n{} {} is deprecated.\n".format(object_type, name)
    msg += "It will be removed in a future version.\n"
    if use_instead is not None:
        msg += "Use {} instead.\n".format(use_instead)
    warnings.warn(msg, DeprecationWarning)
    return msg


def is_bson_valid(value, allow_nonstring_keys=False) -> Tuple[bool, str]:
    """
    returns validity (bool) and a hint (str)
    """
    if isinstance(value, (int, float, str)):
        return True, ''

    if value is None:
        return True, ''

    if isinstance(value, dict):
        for k, v in value.items():
            # Key check
            if isinstance(k, str):
                if '.' in k:
                    log = 'key {} of dict is a string containing a .,' \
                          ' which is forbidden'
                    return False, log.format(k)
            elif isinstance(k, float):
                log = 'key {} of dict is a float, which is forbidden'
                return False, log.format(k)
            elif isinstance(k, int):
                if not allow_nonstring_keys:
                    log = 'key {} of dict is an unsuported type {},' \
                          ' use allow_nonstring_keys=True to allow'
                    return False, log.format(k, type(k))
            else:
                log = 'key {} of dict is an unsuported type {}'
                return False, log.format(k, type(k))

            # Value Check
            v_valid, hint = is_bson_valid(
                value=v, allow_nonstring_keys=allow_nonstring_keys
            )
            if not v_valid:
                return False, hint

    elif is_sequence(value):
        for v in value:
            valid, hint = is_bson_valid(
                value=v, allow_nonstring_keys=allow_nonstring_keys
            )
            if not valid:
                return valid, hint
    else:
        return False, 'Unrecognized type: {}'.format(type(value))
    return True, ''


class DessiaObject:
    """
    Base class for Dessia's platform compatible objects.
    Gathers generic methods and attributes

    :cvar bool _standalone_in_db:
        Indicates wether class objects should be independant
        in database or not.
        If False, object will only exist inside its parent.
    :cvar bool _eq_is_data_eq:
        Indicates which type of equality check is used:
        strict equality or equality based on data. If False, Python's
        object __eq__ method is used (ie. strict),
        else, user custom data_eq is used (ie. data)
    :cvar List[str] _non_serializable_attributes:
        [Advanced] List of instance attributes that should not
        be part of serialization with to_dict method. These will not
        be displayed in platform object tree, for instance.
    :cvar List[str] _non_data_eq_attributes:
        [Advanced] List of instance attributes that should not
        be part of equality check with data__eq__ method
        (if _eq_is_data_eq is True).
    :cvar List[str] _non_data_hash_attributes:
        [Advanced] List of instance attributes that should not
        be part of hash computation with data__hash__ method
        (if _eq_is_data_eq is True).
    :cvar List[str] _ordered_attributes:
        Documentation not available yet.
    :cvar List[str] _titled_attributes:
        Documentation not available yet.
    :cvar List[str] _init_variables:
        Documentation not available yet.
    :cvar List[str] _export_formats:
        List of all available export formats. Class must define a
        export_[format] for each format in _export_formats
    :cvar List[str] _allowed_methods:
        List of all methods that are runnable from platform.
    :cvar List[str] _whitelist_attributes:
        Documentation not available yet.
    :cvar List[str] _whitelist_attributes: List[str]


    :ivar str name: Name of object.
    :ivar Any **kwargs: Additionnal user metadata
    """
    _standalone_in_db = False
    _non_serializable_attributes = []
    _non_editable_attributes = []
    _non_data_eq_attributes = ['name']
    _non_data_hash_attributes = ['name']
    _ordered_attributes = []
    _titled_attributes = []
    _eq_is_data_eq = True

    _init_variables = None
    _allowed_methods = []
    _whitelist_attributes = []

    def __init__(self, name: str = '', **kwargs):
        self.name = name
        for property_name, property_value in kwargs.items():
            setattr(self, property_name, property_value)

    def __hash__(self):
        if self._eq_is_data_eq:
            return self._data_hash()
        else:
            return object.__hash__(self)

    def __eq__(self, other_object):
        if self._eq_is_data_eq:
            if self.__class__.__name__ != other_object.__class__.__name__:
                return False
            if self._data_hash() != other_object._data_hash():
                return False
            return self._data_eq(other_object)
        return object.__eq__(self, other_object)

    def _data_eq(self, other_object):
        return data_eq(self, other_object)

    def _data_hash(self):
        hash_ = 0
        forbidden_keys = (self._non_data_eq_attributes
                          + self._non_data_hash_attributes
                          + ['package_version', 'name'])
        for key, value in self._serializable_dict().items():
            if key not in forbidden_keys:
                if is_sequence(value):
                    hash_ += list_hash(value)
                elif isinstance(value, dict):
                    hash_ += dict_hash(value)
                elif isinstance(value, str):
                    hash_ += sum([ord(v) for v in value])
                else:
                    hash_ += hash(value)
        return int(hash_ % 1e5)

    def _data_diff(self, other_object):
        """
        Make a diff between two objects
        returns: different values, missing keys in other object
        """
        # return diff(self, other_object)
        return diff(self, other_object)

    @property
    def full_classname(self):
        return full_classname(self)

    def base_dict(self):
        package_name = self.__module__.split('.')[0]
        if package_name in sys.modules:
            package = sys.modules[package_name]
            if hasattr(package, '__version__'):
                package_version = package.__version__
            else:
                package_version = None
        else:
            package_version = None

        object_class = self.__module__ + '.' + self.__class__.__name__
        dict_ = {'name': self.name, 'object_class': object_class}
        if package_version:
            dict_['package_version'] = package_version
        return dict_

    def _serializable_dict(self):

        dict_ = {k: v for k, v in self.__dict__.items()
                 if k not in self._non_serializable_attributes
                 and not k.startswith('_')}
        return dict_

    def to_dict(self, use_pointers=False, memo=None, path: str = '#') -> JsonSerializable:
        """
        Generic to_dict method
        """
        if memo is None:
            memo = {}

        # Default to dict
        serialized_dict = self.base_dict()
        dict_ = self._serializable_dict()
        if use_pointers:
            serialized_dict.update(serialize_dict_with_pointers(dict_, memo, path)[0])
        else:
            serialized_dict.update(serialize_dict(dict_))

        return serialized_dict

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable,
                       force_generic: bool = False,
                       global_dict=None,
                       pointers_memo: Dict[str, Any] = None) -> 'DessiaObject':
        """
        Generic dict_to_object method
        """
        # if hasattr(cls, 'DictToObject'):
        #     deprecation_warning(name='DictToObject', object_type='Function',
        #                         use_instead='dict_to_object')
        #     return cls.DictToObject(dict_)

        if cls is not DessiaObject:
            obj = dict_to_object(dict_=dict_, class_=cls,
                                 force_generic=force_generic,
                                 global_dict=global_dict,
                                 pointers_memo=pointers_memo)
            return obj
        elif 'object_class' in dict_:
            obj = dict_to_object(dict_=dict_, force_generic=force_generic,
                                 global_dict=global_dict,
                                 pointers_memo=pointers_memo)
            return obj
        else:
            # Using default
            # TODO: use jsonschema
            raise NotImplementedError('No object_class in dict')

    @classmethod
    def base_jsonschema(cls):
        jsonschema = deepcopy(JSONSCHEMA_HEADER)
        jsonschema['properties']['name'] = {
            'type': 'string',
            "title": "Object Name",
            "description": "Object name",
            "editable": True,
            "default_value": "Object Name"
        }
        return jsonschema

    @classmethod
    def jsonschema(cls):
        if hasattr(cls, '_jsonschema'):
            _jsonschema = cls._jsonschema
            return _jsonschema

        # Get __init__ method and its annotations
        init = cls.__init__
        if cls._init_variables is None:
            annotations = get_type_hints(init)
        else:
            annotations = cls._init_variables

        # Get ordered variables
        if cls._ordered_attributes:
            ordered_attributes = cls._ordered_attributes
        else:
            ordered_attributes = list(annotations.keys())

        unordered_count = 0

        # Parse docstring
        try:
            parsed_docstring = parse_docstring(cls)
        except Exception:
            parsed_docstring = {
                'description': 'Docstring parsing failed',
                'attributes': {}
            }
        parsed_attributes = parsed_docstring['attributes']

        # Initialize jsonschema
        _jsonschema = deepcopy(JSONSCHEMA_HEADER)

        required_arguments, default_arguments = inspect_arguments(method=init,
                                                                  merge=False)
        _jsonschema['required'] = required_arguments
        _jsonschema['standalone_in_db'] = cls._standalone_in_db
        _jsonschema['description'] = parsed_docstring['description']
        _jsonschema['python_typing'] = str(cls)

        # Set jsonschema
        for annotation in annotations.items():
            name = annotation[0]
            if name in ordered_attributes:
                order = ordered_attributes.index(name)
            else:
                order = len(ordered_attributes) + unordered_count
                unordered_count += 1
            if name in cls._titled_attributes:
                title = cls._titled_attributes[name]
            else:
                title = None

            if name != 'return':
                editable = name not in cls._non_editable_attributes
                annotation_type = type_from_annotation(annotation[1], cls)
                annotation = (annotation[0], annotation_type)
                jss_elt = jsonschema_from_annotation(
                    annotation=annotation, jsonschema_element={},
                    order=order, editable=editable, title=title
                )
                if name in parsed_attributes:
                    description = parsed_attributes[name]['desc']
                    jss_elt[name]['description'] = description
                _jsonschema['properties'].update(jss_elt)
                if name in default_arguments.keys():
                    default = set_default_value(_jsonschema['properties'],
                                                name,
                                                default_arguments[name])
                    _jsonschema['properties'].update(default)

        _jsonschema['classes'] = [cls.__module__ + '.' + cls.__name__]
        _jsonschema['whitelist_attributes'] = cls._whitelist_attributes
        return _jsonschema

    @property
    def _method_jsonschemas(self):
        """
        Generates dynamic jsonschemas for methods of class
        """
        jsonschemas = {}
        class_ = self.__class__

        # TOCHECK Backward compatibility. Will need to be changed
        if hasattr(class_, '_dessia_methods'):
            allowed_methods = class_._dessia_methods
        else:
            allowed_methods = class_._allowed_methods

        valid_method_names = [m for m in dir(class_)
                              if not m.startswith('_')
                              and m in allowed_methods]

        for method_name in valid_method_names:
            method = getattr(class_, method_name)

            if not isinstance(method, property):
                required_args, default_args = inspect_arguments(method=method,
                                                                merge=False)
                annotations = get_type_hints(method)
                if annotations:
                    jsonschemas[method_name] = deepcopy(JSONSCHEMA_HEADER)
                    jsonschemas[method_name]['required'] = []
                    jsonschemas[method_name]['method'] = True
                    for i, annotation in enumerate(annotations.items()):
                        # TOCHECK Not actually ordered
                        argname = annotation[0]
                        if argname not in _FORBIDDEN_ARGNAMES:
                            if argname in required_args:
                                jsonschemas[method_name]['required'].append(
                                    str(i))
                            jsonschema_element = \
                                jsonschema_from_annotation(annotation, {}, i)[
                                    argname]

                            jsonschemas[method_name]['properties'][
                                str(i)] = jsonschema_element
                            if argname in default_args.keys():
                                default = set_default_value(
                                    jsonschemas[method_name]['properties'],
                                    str(i),
                                    default_args[argname])
                                jsonschemas[method_name]['properties'].update(
                                    default)
        return jsonschemas

    def method_dict(self, method_name=None, method_jsonschema=None):
        if method_name is None and method_jsonschema is None:
            msg = 'No method name not jsonschema provided'
            raise NotImplementedError(msg)

        if method_name is not None and method_jsonschema is None:
            method_jsonschema = self._method_jsonschemas[method_name]

        dict_ = default_dict(method_jsonschema)
        return dict_

    def dict_to_arguments(self, dict_, method):
        method_object = getattr(self, method)
        args_specs = inspect.getfullargspec(method_object)
        allowed_args = args_specs.args[1:]

        arguments = {}
        for i, arg in enumerate(allowed_args):
            if str(i) in dict_:
                arg_specs = args_specs.annotations[arg]
                value = dict_[str(i)]
                try:
                    deserialized_value = deserialize_argument(arg_specs, value)
                except TypeError:
                    msg = 'Error in deserialisation of value: '
                    msg += '{} of expected type {}'.format(value, arg_specs)
                    raise TypeError(msg)
                arguments[arg] = deserialized_value
        return arguments

    def save_to_file(self, filepath, indent=2):
        if isinstance(filepath, str):
            if not filepath.endswith('.json'):
                filepath += '.json'
                print('Changing name to {}'.format(filepath))
            file = open(filepath, 'w')
        else:
            file = filepath

        try:
            dict_ = self.to_dict(use_pointers=True)
        except TypeError:
            dict_ = self.to_dict()

        json.dump(dict_, file, indent=indent)

        if isinstance(filepath, str):
            file.close()

    @classmethod
    def load_from_file(cls, filepath):
        if isinstance(filepath, str):
            with open(filepath, 'r') as file:
                dict_ = json.load(file)
        else:
            dict_ = json.loads(filepath.read().decode('utf-8'))
        return cls.dict_to_object(dict_)

    def is_valid(self):
        return True

    def copy(self, deep=True, memo=None):
        if deep:
            return self.__deepcopy__(memo=memo)
        else:
            return self.__copy__()

    def __copy__(self):
        """
        Generic copy use inits of objects
        """
        class_argspec = inspect.getfullargspec(self.__class__)
        dict_ = {}
        for arg in class_argspec.args:
            if arg != 'self':
                value = self.__dict__[arg]

                if hasattr(value, '__copy__'):
                    dict_[arg] = value.__copy__()
                else:
                    dict_[arg] = value
        return self.__class__(**dict_)

    def __deepcopy__(self, memo=None):
        """
        Generic deep copy use inits of objects
        """
        class_argspec = inspect.getfullargspec(self.__class__)
        if memo is None:
            memo = {}
        dict_ = {}
        for arg in class_argspec.args:
            if arg != 'self':
                dict_[arg] = deepcopy_value(getattr(self, arg), memo=memo)
        return self.__class__(**dict_)

    def volmdlr_volume_model(self, **kwargs):
        if hasattr(self, 'volmdlr_primitives'):
            import volmdlr as vm  # !!! Avoid circular imports, is this OK ?
            if hasattr(self, 'volmdlr_primitives_step_frames'):
                return vm.core.MovingVolumeModel(
                    self.volmdlr_primitives(**kwargs),
                    self.volmdlr_primitives_step_frames(**kwargs)
                )
            return vm.core.VolumeModel(self.volmdlr_primitives(**kwargs))
        msg = 'Object of type {} does not implement volmdlr_primitives'
        raise NotImplementedError(msg.format(self.__class__.__name__))

    def plot(self, **kwargs):
        """

        """
        if hasattr(self, 'plot_data'):
            import plot_data
            for data in self.plot_data(**kwargs):
                plot_data.plot_canvas(plot_data_object=data,
                                      canvas_id='canvas',
                                      width=1400, height=900,
                                      debug_mode=False)
        else:
            msg = 'Class {} does not implement a plot_data method' \
                  ' to define what to plot'
            raise NotImplementedError(msg.format(self.__class__.__name__))

    def mpl_plot(self, **kwargs):
        axs = []
        if hasattr(self, 'plot_data'):
            try:
                plot_datas = self.plot_data(**kwargs)
            except TypeError as error:
                raise TypeError('{}.{}'.format(self.__class__.__name__, error))
            for data in plot_datas:
                if hasattr(data, 'mpl_plot'):
                    ax = data.mpl_plot()
                    axs.append(ax)
        else:
            msg = 'Class {} does not implement a plot_data method' \
                  'to define what to plot'
            raise NotImplementedError(msg.format(self.__class__.__name__))

        return axs

    def babylonjs(self, use_cdn=True, debug=False, **kwargs):
        self.volmdlr_volume_model(**kwargs).babylonjs(use_cdn=use_cdn,
                                                      debug=debug)

    def save_babylonjs_to_file(self, filename: str = None, use_cdn: bool = True,
                               debug: bool = False, **kwargs):
        self.volmdlr_volume_model(**kwargs).save_babylonjs_to_file(filename=filename,
                                                                   use_cdn=use_cdn,
                                                                   debug=debug)

    def _displays(self, **kwargs) -> List[JsonSerializable]:
        if hasattr(self, '_display_angular'):
            # Retro-compatibility
            deprecation_warning(name='_display_angular', object_type='method',
                                use_instead='display_angular')
            return self._display_angular(**kwargs)

        if 'reference_path' in kwargs:
            reference_path = kwargs['reference_path']
        else:
            reference_path = ''
        displays = []
        if hasattr(self, 'babylon_data'):
            display_ = DisplayObject(type_='cad', data=self.babylon_data(),
                                     reference_path=reference_path)
            displays.append(display_.to_dict())
        elif hasattr(self, 'volmdlr_primitives')\
                or (self.__class__.volmdlr_volume_model
                    is not DessiaObject.volmdlr_volume_model):
            model = self.volmdlr_volume_model()
            display_ = DisplayObject(type_='cad', data=model.babylon_data(),
                                     reference_path=reference_path)
            displays.append(display_.to_dict())
        if hasattr(self, 'plot_data'):
            plot_data = self.plot_data()
            if is_sequence(plot_data):
                for plot in plot_data:
                    display_ = DisplayObject(type_='plot_data', data=plot,
                                             reference_path=reference_path)
                    displays.append(display_.to_dict())
            else:
                msg = 'plot_data must return a sequence. Found {}'
                raise ValueError(msg.format(type(plot_data)))
        if hasattr(self, 'to_markdown'):
            markdown = self.to_markdown()
            display_ = DisplayObject(type_='markdown', data=markdown,
                                     reference_path=reference_path)
            displays.append(display_.to_dict())
        return displays

    def _check_platform(self):
        """
        Reproduce lifecycle on platform (serialization, display)
        """
        try:
            dict_ = self.to_dict(use_pointers=True)
        except TypeError:
            dict_ = self.to_dict()
        json_dict = json.dumps(dict_)
        decoded_json = json.loads(json_dict)
        deserialized_object = self.dict_to_object(decoded_json)
        assert deserialized_object._data_eq(self)
        valid, hint = is_bson_valid(stringify_dict_keys(dict_))
        if not valid:
            raise ValueError(hint)
        json.dumps(self._displays())
        json.dumps(self._method_jsonschemas)

    def to_xlsx(self, filepath):
        writer = XLSXWriter(self)
        writer.save_to_file(filepath)

    def to_step(self, filepath):
        """
        filepath can be a str or an io.StringIO
        """
        return self.volmdlr_volume_model().to_step(filepath=filepath)

    def to_stl(self, filepath):
        """
        filepath can be a str or an io.StringIO
        """
        return self.volmdlr_volume_model().to_stl(filepath=filepath)

    def _export_formats(self):
        formats = [('json', 'save_to_file', True),
                   ('xlsx', 'to_xlsx', False)]
        if hasattr(self, 'volmdlr_primitives'):
            formats.append(('step', 'to_step', True))
            formats.append(('stl', 'to_stl', False))
        return formats


class Catalog(DessiaObject):
    def __init__(self, objects: List[DessiaObject], name: str = ''):
        self.objects = objects
        DessiaObject.__init__(self, name=name)


class DisplayObject(DessiaObject):
    def __init__(self, type_: str,
                 data: Union[JsonSerializable, DessiaObject],
                 reference_path: str = '', name: str = ''):
        if type_ == 'markdown':
            data = inspect.cleandoc(data)
        self.type_ = type_
        self.data = data

        self.reference_path = reference_path
        DessiaObject.__init__(self, name=name)


class Parameter(DessiaObject):
    def __init__(self, lower_bound, upper_bound, periodicity=None, name=''):
        DessiaObject.__init__(self, name=name,
                              lower_bound=lower_bound,
                              upper_bound=upper_bound,
                              periodicity=periodicity)

    def random_value(self):
        return random.uniform(self.lower_bound, self.upper_bound)

    def are_values_equal(self, value1, value2, tol=1e-2):
        if self.periodicity is not None:
            value1 = value1 % self.periodicity
            value2 = value2 % self.periodicity

        return math.isclose(value1, value2, abs_tol=tol)

    def normalize(self, value):
        normalized_value = (value - self.lower_bound) / (self.upper_bound - self.lower_bound)
        return normalized_value

    def original_value(self, normalized_value):
        value = normalized_value * (self.upper_bound - self.lower_bound) + self.lower_bound
        return value

    def optimizer_bounds(self):
        if self.periodicity is not None:
            return (self.lower_bound - 0.5 * self.periodicity,
                    self.upper_bound + 0.5 * self.periodicity)


class ParameterSet(DessiaObject):
    def __init__(self, values, name=''):
        self.values = values

        DessiaObject.__init__(self, name=name)

    @property
    def parameters(self):
        parameters = [Parameter(min(v), max(v), name=k)
                      for k, v in self.values.items()]
        return parameters

    @property
    def means(self):
        means = {k: sum(v) / len(v) for k, v in self.values.items()}
        return means


class DessiaFilter(DessiaObject):
    def __init__(self, attribute: str, operator: str,
                 bound: float, name: str = ''):
        self.attribute = attribute
        self.operator = operator
        self.bound = bound

        DessiaObject.__init__(self, name=name)

    def __hash__(self):
        hash_ = len(self.attribute)
        hash_ += hash(self.operator)
        hash_ += hash(self.bound)
        return int(hash_)

    def __eq__(self, other: 'DessiaFilter'):
        same_attr = self.attribute == other.attribute
        same_op = self.operator == other.operator
        same_bound = self.bound == other.bound
        return same_attr and same_op and same_bound


class Evolution(DessiaObject):
    """
    Defines a generic evolution

    :param evolution: float list
    :type evolution: list
    """
    _non_data_eq_attributes = ['name']
    _non_data_hash_attributes = ['name']
    _generic_eq = True

    def __init__(self, evolution: List[float] = None, name: str = ''):
        if evolution is None:
            evolution = []
        self.evolution = evolution

        DessiaObject.__init__(self, name=name)

    def _displays(self):
        displays = [{'angular_component': 'app-evolution1d',
                     'table_show': False,
                     'evolution': [self.evolution],
                     'label_y': ['evolution']}]
        return displays

    def update(self, evolution):
        """
        Update the evolution list
        """
        self.evolution = evolution


class CombinationEvolution(DessiaObject):
    _non_data_eq_attributes = ['name']
    _non_data_hash_attributes = ['name']
    _generic_eq = True

    def __init__(self, evolution1: List[Evolution],
                 evolution2: List[Evolution], title1: str = 'x',
                 title2: str = 'y', name: str = ''):

        self.evolution1 = evolution1
        self.evolution2 = evolution2

        self.x_, self.y_ = self.genere_xy()

        self.title1 = title1
        self.title2 = title2

        DessiaObject.__init__(self, name=name)

    def _displays(self):
        displays = [{
            'angular_component': 'app-evolution2d-combination-evolution',
            'table_show': False,
            'evolution_x': [self.x_], 'label_x': ['title1'],
            'evolution_y': [self.y_], 'label_y': ['title2']
        }]
        return displays

    def update(self, evol1, evol2):
        """
        Update the CombinationEvolution object

        :param evol1: list
        :param evol2: list
        """
        for evolution, ev1 in zip(self.evolution1, evol1):
            evolution.update(ev1)
        for evolution, ev2 in zip(self.evolution2, evol2):
            evolution.update(ev2)
        self.x_, self.y_ = self.genere_xy()

    def genere_xy(self):
        x, y = [], []
        for evol in self.evolution1:
            x = x + evol.evolution
        for evol in self.evolution2:
            y = y + evol.evolution
        return x, y


def dict_merge(old_dct, merge_dct, add_keys=True, extend_lists=True):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.

    This version will return a copy of the dictionary and leave the original
    arguments untouched.

    The optional argument ``add_keys``, determines whether keys which are
    present in ``merge_dct`` but not ``dct`` should be included in the
    new dict.

    Args:
        old_dct (dict) onto which the merge is executed
        merge_dct (dict): dct merged into dct
        add_keys (bool): whether to add new keys
        extend_lists (bool) : wether to extend lists if keys are updated
                              and value is a list

    Returns:
        dict: updated dict
    """
    dct = deepcopy(old_dct)
    if not add_keys:
        merge_dct = {k: merge_dct[k]
                     for k in set(dct).intersection(set(merge_dct))}

    for key, value in merge_dct.items():
        if isinstance(dct.get(key), dict)\
                and isinstance(value, collections.Mapping):
            dct[key] = dict_merge(dct[key], merge_dct[key],
                                  add_keys=add_keys, extend_lists=extend_lists)
        elif isinstance(dct.get(key), list) and extend_lists:
            dct[key].extend(value)
        else:
            dct[key] = value

    return dct


def stringify_dict_keys(obj):
    if isinstance(obj, (list, tuple)):
        new_obj = []
        for elt in obj:
            new_obj.append(stringify_dict_keys(elt))

    elif isinstance(obj, dict):
        new_obj = {}
        for key, value in obj.items():
            new_obj[str(key)] = stringify_dict_keys(value)
    else:
        return obj
    return new_obj


def serialize_dict(dict_):
    serialized_dict = {}
    for key, value in dict_.items():
        if hasattr(value, 'to_dict'):
            # try:
            #     serialized_value = value.to_dict()
            # except TypeError:
            #     # case of a class as an
            serialized_value = value.to_dict()
        elif isinstance(value, dict):
            serialized_value = serialize_dict(value)
        elif isinstance(value, (list, tuple)):
            serialized_value = serialize_sequence(value)
        else:
            if not is_jsonable(value):
                msg = 'Attribute {} of value {} is not json serializable'
                raise dessia_common.errors.SerializationError(msg.format(key, value))
            serialized_value = value
        serialized_dict[key] = serialized_value
    return serialized_dict


def serialize_sequence(seq):
    serialized_sequence = []
    for value in seq:
        if hasattr(value, 'to_dict'):
            serialized_sequence.append(value.to_dict())
        elif isinstance(value, dict):
            serialized_sequence.append(serialize_dict(value))
        elif isinstance(value, (list, tuple)):
            serialized_sequence.append(serialize_sequence(value))
        else:
            serialized_sequence.append(value)
    return serialized_sequence


def serialize(deserialized_element):
    if isinstance(deserialized_element, DessiaObject):
        serialized = deserialized_element.to_dict()
    elif isinstance(deserialized_element, dict):
        serialized = serialize_dict(deserialized_element)
    elif is_sequence(deserialized_element):
        serialized = serialize_sequence(deserialized_element)
    else:
        serialized = deserialized_element
    return serialized


def list_hash(list_):
    hash_ = 0
    for element in list_:
        if is_sequence(element):
            hash_ += list_hash(element)
        elif isinstance(element, dict):
            hash_ += dict_hash(element)
        elif isinstance(element, str):
            hash_ += sum([ord(e) for e in element])
        else:
            hash_ += hash(element)
    return hash_


def dict_hash(dict_):
    hash_ = 0
    for key, value in dict_.items():
        if is_sequence(value):
            hash_ += list_hash(value)
        elif isinstance(value, dict):
            hash_ += dict_hash(value)
        else:
            hash_ += hash(key) + hash(value)
    return hash_


def getdeepattr(obj, attr):
    return reduce(getattr, [obj] + attr.split('.'))


def enhanced_deep_attr(obj, sequence):
    """
    Get deep attribute where Objects, Dicts and Lists
    can be found in recursion.

    :param obj: Parent object in which recursively find attribute
                represented by sequence
    :param sequence: List of strings and integers that represents
                     path to deep attribute.
    :return: Value of deep attribute
    """
    if isinstance(sequence, str):
        # Sequence is a string and not a sequence of deep attributes
        if '/' in sequence:
            # Is deep attribute reference
            sequence = deepattr_to_sequence(sequence)
            return enhanced_deep_attr(obj=obj, sequence=sequence)
        # Is direct attribute
        return enhanced_get_attr(obj=obj, attr=sequence)

    # Get direct attrivute
    subobj = enhanced_get_attr(obj=obj, attr=sequence[0])
    if len(sequence) > 1:
        # Recursively get deep attributes
        subobj = enhanced_deep_attr(obj=subobj, sequence=sequence[1:])
    return subobj


def enhanced_get_attr(obj, attr):
    """
    Safely get attribute in obj.
    Obj can be of Object, Dict, or List type

    :param obj: Parent object in which find given attribute
    :param attr: String or integer that represents
                      name or index of attribute
    :return: Value of attribute
    """
    try:
        return getattr(obj, attr)
    except (TypeError, AttributeError):
        track = tb.format_exc()
        try:
            return obj[attr]
        except TypeError:
            classname = obj.__class__.__name__
            msg = "'{}' object has no attribute '{}'.".format(classname, attr)
            track += tb.format_exc()
            raise dessia_common.errors.DeepAttributeError(message=msg,
                                                          traceback_=track)


def concatenate_attributes(prefix, suffix, type_: str = 'str'):
    wrong_prefix_format = 'Attribute prefix is wrongly formatted.'
    wrong_prefix_format += 'Is of type {}. Should be str or list'
    if type_ == 'str':
        if isinstance(prefix, str):
            return prefix + '/' + str(suffix)
        elif is_sequence(prefix):
            return sequence_to_deepattr(prefix) + '/' + str(suffix)
        else:
            raise TypeError(wrong_prefix_format.format(type(prefix)))
    elif type_ == 'sequence':
        if isinstance(prefix, str):
            return [prefix, suffix]
        elif is_sequence(prefix):
            return prefix + [suffix]
        else:
            raise TypeError(wrong_prefix_format.format(type(prefix)))
    else:
        wrong_concat_type = 'Type {} for concatenation is not supported.'
        wrong_concat_type += 'Should be "str" or "sequence"'
        raise ValueError(wrong_concat_type.format(type_))


def deepattr_to_sequence(deepattr: str):
    sequence = deepattr.split('/')
    healed_sequence = []
    for i, attribute in enumerate(sequence):
        try:
            healed_sequence.append(int(attribute))
        except ValueError:
            healed_sequence.append(attribute)
    return healed_sequence


def sequence_to_deepattr(sequence):
    healed_sequence = [str(attr) if isinstance(attr, int) else attr
                       for attr in sequence]
    return '/'.join(healed_sequence)


def is_bounded(filter_: DessiaFilter, value: float):
    bounded = True
    operator = filter_.operator
    bound = filter_.bound

    if operator == 'lte' and value > bound:
        bounded = False
    if operator == 'gte' and value < bound:
        bounded = False

    if operator == 'lt' and value >= bound:
        bounded = False
    if operator == 'gt' and value <= bound:
        bounded = False

    if operator == 'eq' and value != bound:
        bounded = False
    return bounded


def type_from_annotation(type_, module):
    """
    Clean up a proposed type if there are stringified
    """
    if isinstance(type_, str):
        # Evaluating types
        if type_ in TYPES_FROM_STRING:
            type_ = TYPES_FROM_STRING[type_]
        else:
            # Evaluating
            type_ = getattr(import_module(module), type_)
    return type_


def jsonschema_from_annotation(annotation, jsonschema_element,
                               order, editable=None, title=None):
    key, typing_ = annotation
    if isinstance(typing_, str):
        raise ValueError

    if title is None:
        title = prettyname(key)
    if editable is None:
        editable = key not in ['return']

    # Compute base entries
    jsonschema_element[key] = {'title': title, 'editable': editable,
                               'order': order,
                               'python_typing': serialize_typing(typing_)}

    if typing_ in TYPING_EQUIVALENCES.keys():
        # Python Built-in type
        jsonschema_element[key]['type'] = TYPING_EQUIVALENCES[typing_]

    elif is_typing(typing_):
        origin = get_origin(typing_)
        args = get_args(typing_)
        if origin is Union:
            if len(args) == 2 and type(None) in args:
                # This is a false Union => Is a default value set to None
                ann = (key, args[0])
                jsonschema_element = jsonschema_from_annotation(
                    annotation=ann, jsonschema_element=jsonschema_element,
                    order=order, editable=editable, title=title
                )
            else:
                # Types union
                classnames = [full_classname(object_=a, compute_for='class')
                              for a in args]

                standalone_args = [a._standalone_in_db for a in args]
                if all(standalone_args):
                    standalone = True
                elif not any(standalone_args):
                    standalone = False
                else:
                    msg = "standalone_in_db values for type '{}'" \
                          " are not consistent"
                    raise ValueError(msg.format(typing_))
                jsonschema_element[key].update({
                    'type': 'object', 'classes': classnames,
                    'standalone_in_db': standalone
                })
        elif origin in [list, collections.Iterator]:
            # Homogenous sequences
            jsonschema_element[key].update(jsonschema_sequence_recursion(
                value=typing_, order=order, title=title, editable=editable
            ))
        elif origin is tuple:
            # Heterogenous sequences (tuples)
            items = []
            for type_ in args:
                items.append({'type': TYPING_EQUIVALENCES[type_]})
            jsonschema_element[key].update({'additionalItems': False,
                                            'type': 'array', 'items': items})
        elif origin is dict:
            # Dynamically created dict structure
            key_type, value_type = args
            if key_type != str:
                # !!! Should we support other types ? Numeric ?
                raise NotImplementedError('Non strings keys not supported')
            jsonschema_element[key].update({
                'type': 'object',
                'patternProperties': {
                    '.*': {
                        'type': TYPING_EQUIVALENCES[value_type]
                    }
                }
            })
        elif origin is Subclass:
            warnings.simplefilter('once', DeprecationWarning)
            msg = "\n\nTyping of attribute '{0}' from class {1} "\
                  "uses Subclass which is deprecated."\
                  "\n\nUse 'InstanceOf[{2}]' instead of 'Subclass[{2}]'.\n"
            arg = args[0].__name__
            warnings.warn(msg.format(key, args[0], arg), DeprecationWarning)
            # Several possible classes that are subclass of another one
            class_ = args[0]
            classname = full_classname(object_=class_, compute_for='class')
            jsonschema_element[key].update({
                'type': 'object', 'instance_of': classname,
                'standalone_in_db': class_._standalone_in_db
            })
        elif origin is InstanceOf:
            # Several possible classes that are subclass of another one
            class_ = args[0]
            classname = full_classname(object_=class_, compute_for='class')
            jsonschema_element[key].update({
                'type': 'object', 'instance_of': classname,
                'standalone_in_db': class_._standalone_in_db
            })
        elif origin is MethodType or origin is ClassMethodType:
            class_type = get_args(typing_)[0]
            classmethod_ = origin is ClassMethodType
            class_jss = jsonschema_from_annotation(
                annotation=('class_', class_type), jsonschema_element={},
                order=order, editable=editable, title='Class'
            )
            jsonschema_element[key].update({
                'type': 'object', 'is_method': True,
                'classmethod_': classmethod_,
                'properties': {
                    'class_': class_jss['class_'],
                    'name': {'type': 'string'}}
            })
        else:
            msg = "Jsonschema computation of typing {} is not implemented"
            raise NotImplementedError(msg.format(typing_))
    elif hasattr(typing_, '__origin__') and typing_.__origin__ is type:
        jsonschema_element[key].update({
            'type': 'object', 'is_class': True,
            'properties': {'name': {'type': 'string'}}
        })
    elif inspect.isclass(typing_) and issubclass(typing_, Measure):
        ann = (key, float)
        jsonschema_element = jsonschema_from_annotation(
            annotation=ann, jsonschema_element=jsonschema_element,
            order=order, editable=editable, title=title
        )
        jsonschema_element[key]['units'] = typing_.units
    elif typing_ is TextIO or typing_ is BinaryIO:
        jsonschema_element[key].update({'type': 'text', 'is_file': True})
    elif typing_ is Any:
        jsonschema_element[key].update({
            'type': 'object',
            'properties': {'.*': '.*'}
        })
    else:
        classname = full_classname(object_=typing_, compute_for='class')
        if inspect.isclass(typing_) and issubclass(typing_, DessiaObject):
            # Dessia custom classes
            jsonschema_element[key].update({
                'type': 'object',
                'standalone_in_db': typing_._standalone_in_db
            })
        else:
            # Statically created dict structure
            jsonschema_element[key].update(static_dict_jsonschema(typing_))
        jsonschema_element[key]['classes'] = [classname]
    return jsonschema_element


def jsonschema_sequence_recursion(value, order: int, title: str = None,
                                  editable: bool = False):
    if title is None:
        title = 'Items'
    jsonschema_element = {'type': 'array', 'order': order,
                          'python_typing': serialize_typing(value)}

    items_type = get_args(value)[0]
    if is_typing(items_type) and get_origin(items_type) is list:
        jss = jsonschema_sequence_recursion(value=items_type, order=0,
                                            title=title, editable=editable)
        jsonschema_element['items'] = jss
    else:
        annotation = ('items', items_type)
        jss = jsonschema_from_annotation(annotation=annotation,
                                         jsonschema_element=jsonschema_element,
                                         order=0, title=title)
        jsonschema_element.update(jss)
    return jsonschema_element


def prettyname(namestr):
    pretty_name = ''
    if namestr:
        strings = namestr.split('_')
        for i, string in enumerate(strings):
            if len(string) > 1:
                pretty_name += string[0].upper() + string[1:]
            else:
                pretty_name += string
            if i < len(strings) - 1:
                pretty_name += ' '
    return pretty_name


def static_dict_jsonschema(typed_dict, title=None):
    warnings.simplefilter('once', DeprecationWarning)
    msg = "\n\nStatic Dict typing is not fully supported.\n" \
          "This will most likely lead to non predictable behavior" \
          " or malfunctionning features. \n" \
          "Define a custom non-standalone class for type '{}'\n\n"
    classname = full_classname(typed_dict, compute_for='class')
    warnings.warn(msg.format(classname), DeprecationWarning)
    jsonschema_element = deepcopy(JSONSCHEMA_HEADER)
    jss_properties = jsonschema_element['properties']

    # Every value is required in a StaticDict
    annotations = get_type_hints(typed_dict)
    jsonschema_element['required'] = list(annotations.keys())

    # TOCHECK : Not actually ordered !
    for i, annotation in enumerate(annotations.items()):
        attribute, typing_ = annotation
        if not is_builtin(typing_):
            msg = "Complex structure as Static dict values is not supported." \
                  "\n Attribute {} got type {} instead of builtin." \
                  "\n Consider creating a custom class for complex structures."
            raise TypeError(msg.format(attribute, typing_))
        jss = jsonschema_from_annotation(annotation=annotation,
                                         jsonschema_element=jss_properties,
                                         order=i, title=title)
        jss_properties.update(jss)
    return jsonschema_element


def set_default_value(jsonschema_element, key, default_value):
    datatype = datatype_from_jsonschema(jsonschema_element[key])
    if default_value is None\
            or datatype in ['builtin', 'heterogeneous_sequence',
                            'static_dict', 'dynamic_dict']:
        jsonschema_element[key]['default_value'] = default_value
    # elif datatype == 'builtin':
    #     jsonschema_element[key]['default_value'] = default_value
    # elif datatype == 'heterogeneous_sequence':
    #     jsonschema_element[key]['default_value'] = default_value
    elif datatype == 'homogeneous_sequence':
        msg = 'Object {} of type {} is not supported as default value'
        type_ = type(default_value)
        raise NotImplementedError(msg.format(default_value, type_))
    elif datatype in ['standalone_object', 'embedded_object',
                      'instance_of', 'union']:
        object_dict = default_value.to_dict()
        jsonschema_element[key]['default_value'] = object_dict
    return jsonschema_element
    # if isinstance(default_value, tuple(TYPING_EQUIVALENCES.keys())) \
    #         or default_value is None:
    #     jsonschema_element[key]['default_value'] = default_value
    # elif is_sequence(default_value):
    #     if datatype == 'heterogeneous_sequence':
    #         jsonschema_element[key]['default_value'] = default_value
    #     else:
    #         msg = 'Object {} of type {} is not supported as default value'
    #         type_ = type(default_value)
    #         raise NotImplementedError(msg.format(default_value, type_))
    # else:
    #     if datatype in ['standalone_object', 'embedded_object',
    #                     'subclass', 'union']:
    #     object_dict = default_value.to_dict()
    #     jsonschema_element[key]['default_value'] = object_dict
    #     else:


def inspect_arguments(method, merge=False):
    # Find default value and required arguments of class construction
    args_specs = inspect.getfullargspec(method)
    nargs = len(args_specs.args) - 1

    if args_specs.defaults is not None:
        ndefault_args = len(args_specs.defaults)
    else:
        ndefault_args = 0

    default_arguments = {}
    arguments = []
    for iargument, argument in enumerate(args_specs.args[1:]):
        if argument not in _FORBIDDEN_ARGNAMES:
            if iargument >= nargs - ndefault_args:
                default_value = args_specs.defaults[ndefault_args - nargs
                                                    + iargument]
                if merge:
                    arguments.append((argument, default_value))
                else:
                    default_arguments[argument] = default_value
            else:
                arguments.append(argument)
    return arguments, default_arguments


def deserialize_argument(type_, argument):
    if argument is None:
        return None

    if is_typing(type_):
        origin = get_origin(type_)
        args = get_args(type_)
        if origin is Union:
            # Check for Union false Positive (Default value = None)
            if len(args) == 2 and type(None) in args:
                return deserialize_argument(type_=args[0], argument=argument)

            # Type union
            classes = list(args)
            instantiated = False
            while instantiated is False:
                # Find the last class in the hierarchy
                hierarchy_lengths = [len(cls.mro()) for cls in classes]
                max_length = max(hierarchy_lengths)
                children_class_index = hierarchy_lengths.index(max_length)
                children_class = classes[children_class_index]
                try:
                    # Try to deserialize
                    # Throws KeyError if we try to put wrong dict into
                    # dict_to_object. This means we try to instantiate
                    # a children class with a parent dict_to_object
                    deserialized_arg = children_class.dict_to_object(argument)

                    # If it succeeds we have the right
                    # class and instantiated object
                    instantiated = True
                except KeyError:
                    # This is not the right class, we should go see the parent
                    classes.remove(children_class)
        elif origin in [list, collections.Iterator]:
            # Homogenous sequences (lists)
            sequence_subtype = args[0]
            deserialized_arg = [deserialize_argument(sequence_subtype, arg)
                                for arg in argument]
            if origin is collections.Iterator:
                deserialized_arg = iter(deserialized_arg)

        elif origin is tuple:
            # Heterogenous sequences (tuples)
            deserialized_arg = tuple([deserialize_argument(t, arg)
                                      for t, arg in zip(args, argument)])
        elif origin is dict:
            # Dynamic dict
            deserialized_arg = argument
        elif origin is InstanceOf:
            classname = args[0]
            object_class = full_classname(object_=classname,
                                          compute_for='class')
            class_ = get_python_class_from_class_name(object_class)
            deserialized_arg = class_.dict_to_object(argument)
        else:
            msg = "Deserialization of typing {} is not implemented"
            raise NotImplementedError(msg.format(type_))
    elif type_ is TextIO:
        deserialized_arg = argument
    elif type_ is BinaryIO:
        # files are supplied as io.BytesIO  which is compatible with : BinaryIO
        deserialized_arg = argument
    else:
        if type_ in TYPING_EQUIVALENCES.keys():
            if isinstance(argument, type_):
                deserialized_arg = argument
            else:
                if isinstance(argument, int) and type_ == float:
                    # Explicit conversion in this case
                    deserialized_arg = float(argument)
                else:
                    msg = 'Given built-in type and argument are incompatible: '
                    msg += '{} and {} in {}'.format(type(argument),
                                                    type_, argument)
                    raise TypeError(msg)
        elif type_ is Any:
            # Any type
            deserialized_arg = argument
        elif inspect.isclass(type_) and issubclass(type_, DessiaObject):
            # Custom classes
            deserialized_arg = type_.dict_to_object(argument)
        else:
            raise TypeError("Deserialization of ype {} is Not Implemented".format(type_))
    return deserialized_arg


# TODO recursive_type and recursive_type functions look weird
def recursive_type(obj):
    if isinstance(obj, tuple(list(TYPING_EQUIVALENCES.keys()) + [dict])):
        type_ = TYPES_STRINGS[type(obj)]
    elif isinstance(obj, DessiaObject):
        type_ = obj.__module__ + '.' + obj.__class__.__name__
    elif isinstance(obj, (list, tuple)):
        type_ = []
        for element in obj:
            type_.append(recursive_type(element))
    elif obj is None:
        type_ = None
    else:
        raise NotImplementedError(obj)
    return type_


def recursive_instantiation(type_, value):
    if type_ in TYPES_STRINGS.values():
        return eval(type_)(value)
    elif isinstance(type_, str):
        class_ = get_python_class_from_class_name(type_)
        if inspect.isclass(class_):
            return class_.dict_to_object(value)
        else:
            raise NotImplementedError
    elif isinstance(type_, (list, tuple)):
        return [recursive_instantiation(t, v) for t, v in zip(type_, value)]
    elif type_ is None:
        return value
    else:
        raise NotImplementedError(type_)


def default_sequence(array_jsonschema):
    if is_sequence(array_jsonschema['items']):
        # Tuple jsonschema
        if 'default_value' in array_jsonschema:
            return array_jsonschema['default_value']
        return [default_dict(v) for v in array_jsonschema['items']]
    return None


def datatype_from_jsonschema(jsonschema):
    if jsonschema['type'] == 'object':
        if 'classes' in jsonschema:
            if len(jsonschema['classes']) > 1:
                return 'union'
            if 'standalone_in_db' in jsonschema:
                if jsonschema['standalone_in_db']:
                    return 'standalone_object'
                return 'embedded_object'
            return 'static_dict'
        if 'instance_of' in jsonschema:
            return 'instance_of'
        if 'patternProperties' in jsonschema:
            return 'dynamic_dict'
        if 'method' in jsonschema and jsonschema['method']:
            return 'embedded_object'
        if 'is_class' in jsonschema and jsonschema['is_class']:
            return 'class'

    elif jsonschema['type'] == 'array':
        if 'additionalItems' in jsonschema\
                and not jsonschema['additionalItems']:
            return 'heterogeneous_sequence'
        return 'homogeneous_sequence'

    elif jsonschema['type'] in ['number', 'string', 'boolean']:
        if 'is_type' in jsonschema and jsonschema['is_type']:
            return 'file'
        return 'builtin'
    return None


def chose_default(jsonschema):
    datatype = datatype_from_jsonschema(jsonschema)
    if datatype in ['heterogeneous_sequence', 'homogeneous_sequence']:
        return default_sequence(jsonschema)
    elif datatype == 'static_dict':
        return default_dict(jsonschema)
    elif datatype in ['standalone_object', 'embedded_object',
                      'instance_of', 'union']:
        if 'default_value' in jsonschema:
            return jsonschema['default_value']
        return None
    else:
        return None


def default_dict(jsonschema):
    dict_ = {}
    datatype = datatype_from_jsonschema(jsonschema)
    if datatype in ['standalone_object', 'embedded_object', 'static_dict']:
        if 'classes' in jsonschema:
            dict_['object_class'] = jsonschema['classes'][0]
        elif 'method' in jsonschema and jsonschema['method']:
            # Method can have no classes in jsonschema
            pass
        else:
            msg = "DessiaObject of type {} must have 'classes' in jsonschema"
            raise ValueError(msg.format(jsonschema['python_typing']))
        for property_, jss in jsonschema['properties'].items():
            if 'default_value' in jss:
                dict_[property_] = jss['default_value']
            else:
                dict_[property_] = chose_default(jss)
    else:
        return None
    return dict_


class ParsedAttribute(TypedDict):
    desc: str
    type_: str
    annotation: str


class ParsedDocstring(TypedDict):
    description: str
    attributes: Dict[str, ParsedAttribute]


def parse_docstring(cls: Type) -> ParsedDocstring:
    """
    Parse docstring of given class. Refer to docs to see how docstrings
    should be built.
    """
    annotations = get_type_hints(cls.__init__)
    docstring = cls.__doc__
    if docstring:
        splitted_docstring = docstring.split(':param ')
        parsed_docstring = {"description": splitted_docstring[0].strip()}
        params = splitted_docstring[1:]
        args = {}
        for param in params:
            splitted_param = param.split(':type ')
            arg = splitted_param[0]
            typestr = splitted_param[1]
            argname, argdesc = arg.split(":", maxsplit=1)
            argtype = typestr.split(argname + ":")[-1]
            annotation = annotations[argname]
            args[argname] = {'desc': argdesc.strip(), 'type_': argtype.strip(),
                             'annotation': str(annotation)}
            # TODO Should be serialize typing ?
        parsed_docstring.update({'attributes': args})
        return parsed_docstring
    return {'description': "", 'attributes': {}}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 19 17:32:48 2021

@author: masfaraud
"""

from .power_simulation import simulation_workflow
from .forms_workflow import workflow_
from .vectored_workflow import vectored_workflow


import webbrowser
import os
import tempfile
from dessia_common.templates import visjs_template
from networkx import DiGraph, Graph, kamada_kawai_layout


def networkx_to_visjs_data(networkx_graph: Graph):
    visjs_data = {'name': networkx_graph.name, 'nodes': [], 'edges': []}

    pos = kamada_kawai_layout(networkx_graph)

    for i, node in enumerate(networkx_graph.nodes):
        node_dict = networkx_graph.nodes[node]
        node_data = {'id': i}

        if node in pos:
            node_data['x'], node_data['y'] = pos[node]

        if 'name' not in node_dict and 'label' not in node_dict:
            if isinstance(node, str):
                node_data['label'] = node
            elif isinstance(node, int):
                node_data['label'] = str(node)
            else:
                node_data['label'] = ''
        elif 'name' in node_dict and 'label' not in node_dict:
            node_data['label'] = node_dict['name']
        else:
            node_data['label'] = node_dict['name'] + node_dict['label']

        if 'shape' not in node_dict:
            node_data['shape'] = 'circular'
        else:
            node_data['shape'] = node_dict['shape']

        if 'color' in node_dict:
            node_data['color'] = node_dict['color']

        visjs_data['nodes'].append(node_data)

    list_nodes = list(networkx_graph.nodes)
    is_digraph = isinstance(networkx_graph, DiGraph)
    print(is_digraph)
    for edge in networkx_graph.edges:
        index1 = list_nodes.index(edge[0])
        index2 = list_nodes.index(edge[1])
        edge_data = {'from': index1,
                     'to': index2,
                     'font': {'align': 'middle'}}

        if is_digraph:
            edge_data['arrows'] = 'to'

        visjs_data['edges'].append(edge_data)

    return visjs_data


def draw_networkx_graph(networkx_graph: Graph):
    visjs_data = networkx_to_visjs_data(networkx_graph)
    s = visjs_template.substitute(**visjs_data)
    with tempfile.NamedTemporaryFile(suffix=".html",
                                     delete=False) as file:
        file.write(bytes(s, 'utf8'))

    # with open('graph_visJS.html', 'wb') as file:
    #     file.write(s.encode('utf-8'))
    webbrowser.open('file://' + os.path.realpath(file.name))
    return file.name
